package abc.ar;

import java.io.Reader;
import java.util.Collection;
import java.util.List;

import polyglot.ast.NodeFactory;
import polyglot.frontend.CupParser;
import polyglot.frontend.FileSource;
import polyglot.frontend.GlobalBarrierPass;
import polyglot.frontend.Job;
import polyglot.frontend.Parser;
import polyglot.frontend.Pass;
import polyglot.frontend.VisitorPass;
import polyglot.frontend.Pass.ID;
import polyglot.lex.Lexer;
import polyglot.types.TypeSystem;
import polyglot.util.ErrorQueue;
import abc.ar.ast.ARNodeFactory;
import abc.ar.ast.ARNodeFactory_c;
import abc.ar.parse.Grm;
import abc.ar.types.ARTypeSystem_c;
import abc.ar.visit.ExecutionPointcutRenamer;
import abc.ar.visit.FeatureNameResolverVisitor;
import abc.ar.visit.NamedAdviceMethodGenerator;
import abc.ar.visit.NamedAdviceMethodRemover;
import abc.ar.visit.NamedAspectMethods;
import abc.ar.visit.NamedBeforeAfterAdviceWeaver;
import abc.ar.visit.PointcutSuperResolver;
import abc.ar.visit.PrintAspectsVisitor;
import abc.ar.visit.RefinedAspectRenamingVisitor;
import abc.ar.visit.RefinementCollectorVisitor;
import abc.ar.visit.RefinementHierachieBuilder;
import abc.ar.visit.RefinementHierachieWeaver;
import abc.aspectj.parse.Lexer_c;
import abc.aspectj.types.AJTypeSystem;
import abc.aspectj.visit.NamePatternReevaluator;

/**
 * Extension information for Extended AspectJ extension.
 * 
 * @author Christian Kaestner and Sven Apel
 */
public class ExtensionInfo extends abc.aspectj.ExtensionInfo {
	public static final Pass.ID RESOLVED_ASPECT_REFINEMENTS = new Pass.ID("resolved-aspect-refinements");

	public static final Pass.ID MARKED_ASPECT_REFINEMENTS1 = new Pass.ID("marked-aspect-refinements1");

	public static final Pass.ID MARKED_ASPECT_REFINEMENTS2 = new Pass.ID("marked-aspect-refinements2");

	public static final Pass.ID MARKED_ASPECT_REFINEMENTSBARRIER = new Pass.ID("marked-aspect-refinementsbarrier");

	public static final Pass.ID MARKED_ASPECT_REFINEMENTSBARRIER2 = new Pass.ID("marked-aspect-refinementsbarrier2");

	public static final Pass.ID TEST_VISITOR = new Pass.ID("test-visitor");

	public static final Pass.ID SET_SUPER_BY_REFINEMENTS = new Pass.ID("refinement-super-setter");

	public static final Pass.ID PRINT_ASPECTS = new Pass.ID("print-aspects");

	public static final Pass.ID RESOLVE_FEATURE_NAMES = new Pass.ID("b1");

	public static final Pass.ID BUILD_REFINEMENT_HIERACHIE = new Pass.ID("b2");

	public static final Pass.ID COLLECT_REFINEMENTS = new Pass.ID("b3");

	public static final Pass.ID B4 = new Pass.ID("b4");

	public static final Pass.ID RENAMED_REFINED_ASPECTS = new Pass.ID("b5");

	public static final Pass.ID WEAVE_REFINED_HIERACHIE = new Pass.ID("b6");

	public static final Pass.ID B7 = new Pass.ID("b7");

	public static final Pass.ID C1 = new Pass.ID("c1");

	public static final Pass.ID B1 = new Pass.ID("bb1");
	public static final Pass.ID B2 = new Pass.ID("bb2");

	public static final Pass.ID REMOVE_NAMED_METHODS = new Pass.ID("d1");

	public static final Pass.ID GENERATE_NAMED_METHODS = new Pass.ID("d2");

	public static final Pass.ID WEAVE_NAMED_BEFOREAFTERADVICE = new Pass.ID("d3");

	public static final Pass.ID RENAME_EXECUTIONPOINTCUT = new Pass.ID("e1");

	public static final Pass.ID R = new Pass.ID("r");

	static {
		// force Topics to load
		Topics t = new Topics();
	}

	public ExtensionInfo(Collection jar_classes, Collection source_files) {
		super(jar_classes, source_files);
	}

	public String compilerName() {
		return "ar";
	}

	public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
		Lexer lexer = new Lexer_c(reader, source.name(), eq);
		Grm grm = new Grm(lexer, ts, nf, eq);
		return new CupParser(grm, source, eq);
	}

	protected NodeFactory createNodeFactory() {
		return new ARNodeFactory_c();
	}

	protected TypeSystem createTypeSystem() {
		return new ARTypeSystem_c();
	}
	

	protected void passes_parse_and_clean(List l, Job job) {
		super.passes_parse_and_clean(l, job);

		int idx = findPassIdx(l, Pass.BUILD_TYPES_ALL);
		if (idx == -1)
			idx = l.size() - 1;

		idx--;

		// first get the feature names for all aspects
		l.add(++idx, new VisitorPass(RESOLVE_FEATURE_NAMES, job, new FeatureNameResolverVisitor(job, ts, (ARNodeFactory) nf)));
		// then collect all refinement and all aspects into a global structure
		l.add(++idx, new VisitorPass(COLLECT_REFINEMENTS, job, new RefinementCollectorVisitor(job, ts, (ARNodeFactory) nf)));
		l.add(++idx, new GlobalBarrierPass(B4, job));
		// calculate the changes to the hierachy
		l.add(++idx, new RefinementHierachieBuilder(BUILD_REFINEMENT_HIERACHIE, job));
		// rename the refined aspects
		l.add(++idx, new VisitorPass(RENAMED_REFINED_ASPECTS, job, new RefinedAspectRenamingVisitor(job, ts, (ARNodeFactory) nf)));
		l.add(++idx, new GlobalBarrierPass(B7, job));
		// weave the hierachie, this makes refined aspects abstract and sets the
		// superclasses for refinements
		l.add(++idx, new VisitorPass(WEAVE_REFINED_HIERACHIE, job, new RefinementHierachieWeaver(job, ts, (ARNodeFactory) nf)));

		// debug only: show all aspects
		// if (Debug.v.refinementFrontentVisitor)
		// l.add(++idx, new VisitorPass(PRINT_ASPECTS, job, new
		// PrintAspectsVisitor(job, ts, (ARNodeFactory) nf)));

	}

	protected void passes_add_members(List l, Job job) {
		l.add(new VisitorPass(REMOVE_NAMED_METHODS, job, new NamedAdviceMethodRemover(job, ts, (ARNodeFactory) nf)));
		l.add(new VisitorPass(GENERATE_NAMED_METHODS, job, new NamedAdviceMethodGenerator(job, (AJTypeSystem) ts, (ARNodeFactory) nf)));
		l.add(new VisitorPass(WEAVE_NAMED_BEFOREAFTERADVICE, job, new NamedBeforeAfterAdviceWeaver(job, ts, (ARNodeFactory) nf)));
		super.passes_add_members(l, job);
	}
	
	protected void passes_disambiguate_all(List l, Job job) {
		// TODO Auto-generated method stub
		super.passes_disambiguate_all(l, job);
		l.add(new VisitorPass(PRINT_ASPECTS, job, new PrintAspectsVisitor(job, ts, (ARNodeFactory) nf)));
	}


	private int findPassIdx(List l, ID passId) {
		for (int i = 0; i < l.size(); i++) {
			Pass pass = (Pass) l.get(i);
			if (pass.id() == passId)
				return i;

		}
		return -1;
	}

	protected void passes_patterns_and_parents(List l, Job job) {

		super.passes_patterns_and_parents(l, job);

		l.add(new VisitorPass(C1, job, new PointcutSuperResolver(job, ts, (ARNodeFactory) nf)));



		int i = findPassIdx(l, CHECK_PACKAGE_NAMES);
		l.remove(i);

	}

	protected void passes_aspectj_transforms(List l, Job job) {
		l.add(new VisitorPass(RENAME_EXECUTIONPOINTCUT, job, new ExecutionPointcutRenamer(job, ts, (ARNodeFactory) nf,this)));
        l.add(new GlobalBarrierPass(B1, job));
        l.add(new NamePatternReevaluator(B2));
		
		super.passes_aspectj_transforms(l, job);

		int i = findPassIdx(l, ASPECT_METHODS);
		l.set(i, new VisitorPass(ASPECT_METHODS, job, new NamedAspectMethods(job, nf, ts)));

	}

}
