package abc.ar.visit;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import polyglot.ast.ClassBody;
import polyglot.ast.ClassDecl;
import polyglot.ast.Formal;
import polyglot.ast.MethodDecl;
import polyglot.ast.Node;
import polyglot.ast.TypeNode;
import polyglot.frontend.Job;
import polyglot.types.ClassType;
import polyglot.types.ParsedClassType;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;
import abc.ar.AbcExtension;
import abc.ar.RefinementManager;
import abc.ar.ast.ARNodeFactory;
import abc.ar.ast.NamedAdviceDecl;
import abc.ar.ast.NamedAdviceSpec;
import abc.aspectj.ExtensionInfo;
import abc.aspectj.ast.AspectBody;
import abc.aspectj.ast.CPEName;
import abc.aspectj.ast.ClassTypeDotId;
import abc.aspectj.ast.ClassnamePatternExpr;
import abc.aspectj.ast.NamePattern;
import abc.aspectj.ast.PointcutDecl;
import abc.aspectj.ast.SimpleNamePattern;
import abc.aspectj.types.AJContext;
import abc.aspectj.visit.PatternMatcher;
import abc.main.Debug;

/**
 * this visitor provides a simple facility to change execution pointcuts on
 * refined aspects. if a pointcut is defined on a aspect that is refined the
 * execution pointcut matches the latest refinement that contains the method
 * (similar to the behaviour with call pointcuts)
 * 
 * this visitor only provides a very simple solution for this problem, only
 * based on the methods name. patterns or different signatures are not
 * supported. TODO a better solution for this should be implemented in the
 * pointcut matcher later
 * 
 * 
 * @author Christian Kaestner and Sven Apel
 */
public class ExecutionPointcutRenamer extends ContextVisitor {

	private RefinementManager refinementManager;

	private ExtensionInfo ext;

	protected ARNodeFactory nf;

	public ExecutionPointcutRenamer(Job job, TypeSystem ts, ARNodeFactory nf, ExtensionInfo ext) {
		super(job, ts, nf);
		this.refinementManager = AbcExtension.v().getRefinementManager();
		this.ext = ext;
		this.nf = nf;
	}

	private String currentMethodName;

	protected NodeVisitor enterCall(Node n) throws SemanticException {
		if (n instanceof ClassTypeDotId) {
			currentMethodName = ((ClassTypeDotId) n).name().getPatternString();
		}

		return super.enterCall(n);
	}

	protected Node leaveCall(Node n) throws SemanticException {
		if (n instanceof ClassTypeDotId) {
			currentMethodName = null;
		}
		if (currentMethodName != null && n instanceof CPEName
				&& ((CPEName) n).getNamePattern() instanceof SimpleNamePattern) {
			CPEName base = (CPEName) n;
			String basePatternString = ((SimpleNamePattern) base.getNamePattern()).getPatternString();
			if (refinementManager.isRefinedAspect(basePatternString)) {

				// for refined aspects: go through the hierachy from topmost
				// to bottom
				// to find the refinement that introduced the last version
				// of the targeted method
				// if found rebuild the ClassTypeDotId object with a new
				// base.

				String feature = refinementManager.getTopMostFeatureName(basePatternString);
				while (feature != null) {
					String aspectName = refinementManager.getNameForAspect(basePatternString, feature);
					ClassType aspectType = getClassTypeForName(aspectName);
					List methodList = aspectType.methodsNamed(currentMethodName);
					if (methodList != null && methodList.size() > 0)
						return rewriteCPEName(base, aspectName,context.currentClass());

					feature = refinementManager.getSuperFeature(basePatternString, feature);
				}
			}
		}

		return super.leaveCall(n);
	}

	private Node rewriteCPEName(CPEName oldBase, String aspectName, ClassType currentClassType) {
		// if pattern already equals aspectname return unchanged
		SimpleNamePattern oldPattern = (SimpleNamePattern) oldBase.getNamePattern();
		String basePatternString = oldPattern.getPatternString();
		if (basePatternString.equals(aspectName))
			return oldBase;

		// else rebuild with new base
		SimpleNamePattern newPattern = nf.SimpleNamePattern(oldPattern.position(), aspectName);
		PatternMatcher.v().computeMatches(newPattern,ext.hierarchy.getClass(currentClassType),new HashSet(),new HashSet());
		CPEName newBase = nf.CPEName(oldBase.position(), newPattern);
		return newBase;
	}

	private ClassType getClassTypeForName(String aspectName) {
		Collection types = ext.hierarchy.getClassTypes();
		for (Iterator i = types.iterator(); i.hasNext();) {
			ClassType ct = (ClassType) i.next();
			if (ct.name().equals(aspectName))
				return ct;
		}
		return null;
	}

}
