package abc.ar.visit;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import polyglot.ast.AmbExpr;
import polyglot.ast.Assign;
import polyglot.ast.Block;
import polyglot.ast.Call;
import polyglot.ast.Eval;
import polyglot.ast.Field;
import polyglot.ast.Formal;
import polyglot.ast.Node;
import polyglot.ast.Special;
import polyglot.frontend.Job;
import polyglot.types.FieldInstance;
import polyglot.types.MethodInstance;
import polyglot.types.ParsedClassType;
import polyglot.types.SemanticException;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import polyglot.visit.ContextVisitor;
import polyglot.visit.NodeVisitor;
import abc.ar.ast.ARNodeFactory;
import abc.ar.ast.NamedAdviceDecl;
import abc.ar.ast.NamedAdviceSpec;
import abc.ar.ast.NamedAfterReturning;
import abc.ar.ast.NamedAfterThrowing;
import abc.ar.ast.NamedAround;
import abc.aspectj.ast.AdviceFormal;
import abc.aspectj.ast.AdviceSpec;
import abc.aspectj.ast.AspectDecl;
import abc.main.Debug;

/**
 * generates the calls to the named advices (before and after).
 * 
 * this visitor runs after NamedAdviceMethodGenerator where the named methods
 * are created and the method body is moved to the named method
 * 
 * @author Christian Kaestner and Sven Apel
 */
public class NamedBeforeAfterAdviceWeaver extends ContextVisitor {

	private ARNodeFactory nf;

	public NamedBeforeAfterAdviceWeaver(Job job, TypeSystem ts, ARNodeFactory nf) {
		super(job, ts, nf);
		this.nf = nf;
	}

	private AspectDecl currentAspect = null;

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException {
		if (n instanceof AspectDecl)
			currentAspect = (AspectDecl) n;
		return super.enterCall(parent, n);
	}

	protected Node leaveCall(Node n) throws SemanticException {
		if (n instanceof AspectDecl)
			currentAspect = null;
		if (n instanceof NamedAdviceDecl) {
			NamedAdviceDecl a = (NamedAdviceDecl) n;

			AdviceSpec spec = a.getSpec();
			assert spec instanceof NamedAdviceSpec : "Named advice not specified with named spec.";
			NamedAdviceSpec namedSpec = (NamedAdviceSpec) spec;

			// only before and after advice, not around
			if (!(namedSpec instanceof NamedAround)) {

				String name = namedSpec.name();
				Position p = a.position();

				assert currentAspect != null;

				List statements = new ArrayList();
				//set join point to member variable
				if (a.hasJoinPoint())
					statements.add(assignParamToField(p,"thisJoinPoint"));
				if (a.hasJoinPointStaticPart())
					statements.add(assignParamToField(p,"thisJoinPointStaticPart"));
				if (a.hasEnclosingJoinPointStaticPart())
					statements.add(assignParamToField(p,"thisEnclosingJoinPointStaticPart"));
				
				
				
				// parameter for called function
				List args = new ArrayList();
				for (Iterator i = namedSpec.formals().iterator(); i.hasNext();) {
					Formal f = (Formal) i.next();

					AmbExpr arg = nf.AmbExpr(p, f.name());
					args.add(arg);
				}
				// add return value or exception as parameter
				if (namedSpec instanceof NamedAfterReturning || namedSpec instanceof NamedAfterThrowing) {
					AdviceFormal f = namedSpec.returnVal();
					if (f != null) {
						AmbExpr arg = nf.AmbExpr(p, f.name());
						args.add(arg);
					}
				}

				Call expr = nf.Call(p, name, args);
				Eval eval = nf.Eval(p, expr);
				statements.add(eval);
				Block block = nf.Block(p, statements);
				n = a.body(block);

				if (Debug.v().refinementFrontentVisitor)
					System.out.println("Named Advice: " + name + " - " + a);
			}
		}
		return super.leaveCall(n);
	}

	private Eval assignParamToField(Position p, String paramName) {
		Field this_thisJoinPoint = nf.Field(p, nf.Special(p,Special.THIS), paramName);
		AmbExpr param_thisJointPoint = nf.AmbExpr(p, paramName);
		Assign assign = nf.Assign(p, this_thisJoinPoint, Assign.ASSIGN, param_thisJointPoint);
		return nf.Eval(p, assign);
	}

}
